import os

PROJECT_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

#TEST_DATA = os.path.join(PROJECT_ROOT, "data/train_2008011514_data.json")
#TEST_JSON = os.path.join(PROJECT_ROOT, "test/test.json")
USE_CUDA = False
BERT_MODEL = "bert-base-chinese"
PAD = 0
UNK = 1
CLS = 2
SEP = 3
COMMA = 117
LESS_THAN = 133
LARGER_THAN = 135




import torch
import torch.nn as nn

from transformers import BertModel, BertPreTrainedModel

class WSDBertClassifer(BertPreTrainedModel):
    def __init__(self, config, pos=1):
        super(WSDBertClassifer, self).__init__(config)
        self.loss_function = nn.BCEWithLogitsLoss(pos_weight=torch.tensor(pos))
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifer = nn.Linear(config.hidden_size, 1)
        self.sigmoid = nn.Sigmoid()
        self.init_weights()
    
    def forward(self, context, attention_mask = None, token_type_ids = None, labels = None):
        seq_out, pool_out = self.bert(context, attention_mask=attention_mask, 
                                token_type_ids=token_type_ids, return_dict=False)
        pool_out = self.dropout(pool_out)
        logits = self.classifer(pool_out)
        loss = 0
        if labels is not None:
            loss = self.loss_function(logits.squeeze(1), labels)
        logits = self.sigmoid(logits)
        return logits, loss



from enum import Enum, auto
from .cwn_annot_types import CwnAnnotationInfo

class CwnRelationType(Enum):
    holonym = 1
    antonym = 2
    meronym = 3
    hypernym = 4
    hyponym = 5
    variant = 6
    nearsynonym = 7
    paranym = 8
    synonym = 9
    varword = 11
    instance_of = 12
    has_instance = 13        
    generic = -1    
    has_sense = 91
    has_lemma = 92
    has_facet = 93
    is_synset = 94

    def __repr__(self):
        return f"<CwnRelationType: {str(self.name)}>"

    def __eq__(self, other):
        if isinstance(other, CwnRelationType):
            return other.value == self.value
        else:
            return False

    def is_semantic_relation(self):
        return 0 < self.value <= 20

    def inverse(self):
        cls = self.__class__
        inverse_pairs = [
            (cls.has_instance, cls.instance_of),
            (cls.hypernym, cls.hyponym),
            (cls.holonym, cls.meronym)
        ]

        for rel_x, rel_y in inverse_pairs:
            if self == rel_x:
                return rel_y

            if self == rel_y:
                return rel_x
        
        return None

    @staticmethod
    def from_zhLabel(zhlabel):
        label_map = {
            "全體詞": CwnRelationType.holonym,
            "反義詞": CwnRelationType.antonym,
            "部分詞": CwnRelationType.meronym,
            "上位詞": CwnRelationType.hypernym,
            "下位詞": CwnRelationType.hyponym,
            "異體": CwnRelationType.variant,
            "近義詞": CwnRelationType.nearsynonym,
            "類義詞": CwnRelationType.paranym,
            "同義詞": CwnRelationType.synonym,
            "事例": CwnRelationType.has_instance,
            "之事例": CwnRelationType.instance_of,
            "同義詞集": CwnRelationType.is_synset
        }

        return label_map.get(zhlabel, CwnRelationType.generic)


class CwnRelation(CwnAnnotationInfo):
    def __init__(self, eid, cgu, reversed=False):
        edata = cgu.get_edge_data(eid)
        self.cgu = cgu
        self.id = eid
        self.edge_type = edata.get("edge_type", "generic")
        self.annot = {}
        self.reversed = reversed

    def __repr__(self):
        src_id = self.id[0]
        tgt_id = self.id[1]
        if not self.reversed:
            return f"<CwnRelation> {self.edge_type}: {src_id} -> {tgt_id}"
        else:
            return f"<CwnRelation> {self.edge_type}(rev): {tgt_id} <- {src_id}"

    def data(self):
        data_fields = ["edge_type", "annot"]
        data_dict= {
            k: self.__dict__[k] for k in data_fields
        }
        return data_dict

    @property
    def src_id(self):
        return self.id[0]

    @property
    def tgt_id(self):
        return self.id[1]

    @property
    def relation_type(self):
        return self.edge_type

    @relation_type.setter
    def relation_type(self, x):        
        if not isinstance(x, CwnRelationType):
            raise ValueError(f"{x} is not instance of CwnRelationType")
        else:
            self.edge_type = x.name


from enum import Enum, auto
from .cwn_annot_types import CwnAnnotationInfo
from .cwn_relation_types import CwnRelationType
from collections import namedtuple

class CwnNode:
    def __init__(self):
        self.id = None
        self.node_type = None

    def data(self):
        raise NotImplementedError("abstract method: CwnNode.data")

    def __eq__(self, other):
        raise NotImplementedError()

    def __hash__(self):
        raise NotImplementedError()

class CwnGlyph(CwnAnnotationInfo):
    def __init__(self, nid, cgu):
        ndata = cgu.get_node_data(nid)
        self.glyph = ndata.get("glyph", "")
        self.annot = ndata.get("annot", {})

    def __repr__(self):
        return "<CwnLemma: {lemma}_{lemma_sno}>".format(
            **self.__dict__
        )

    def __eq__(self, other):
        if isinstance(other, CwnGlyph):
            return self.glyph == other.glyph
        else:
            return False

    def __hash__(self):
        return hash(self.glyph)

    def data(self):
        data_fields = ["node_type", "glyph"]
        return {
            k: self.__dict__[k] for k in data_fields
        }

class CwnLemma(CwnAnnotationInfo):
    """Class representing a lemma.

    Attributes
    ----------
    senses: list
        a list of senses (:class:`CwnSense <.CwnSense>`) having this
        lemma
    """
    def __init__(self, nid, cgu):
        ndata = cgu.get_node_data(nid)
        self.cgu = cgu
        self.id = nid
        self.node_type = "lemma"
        self.lemma = ndata.get("lemma", "")
        self.lemma_sno = ndata.get("lemma_sno", 1)
        self.zhuyin = ndata.get("zhuyin", "")
        self.annot = ndata.get("annot", {})
        self._senses = None

    def __repr__(self):
        return "<CwnLemma: {lemma}_{lemma_sno}>".format(
            **self.__dict__
        )

    def __eq__(self, other):
        if isinstance(other, CwnLemma):
            return self.lemma == other.lemma and \
                self.zhuyin == other.zhuyin
        else:
            return False

    def __hash__(self):
        return hash((self.lemma, self.zhuyin))

    def data(self):
        """Retrieve all data of this lemma.
        
        Returns
        -------
        dict
            data stored in a dictionary, including the following
            keys: ``node_type``, ``lemma``, ``lemma_sno``, ``zhuyin``,
            ``annot``
        """
        data_fields = ["node_type", "lemma", "lemma_sno", "zhuyin", "annot"]
        return {
            k: self.__dict__[k] for k in data_fields
        }

    @staticmethod
    def from_word(word, cgu):
        """Find lemmas matching search pattern.
        
        Parameters
        ----------
        word : str
            RegEx pattern to search for.
        cgu : CwnBase
            See :class:`CwnBase <CwnGraph.cwn_base.CwnBase>`.
        
        Returns
        -------
        list
            A list of :class:`CwnLemma <.CwnLemma>`.
        """
        return cgu.find_lemma(word)

    @property
    def senses(self):
        if self._senses is None:
            cgu = self.cgu
            sense_nodes = []
            edges = cgu.find_edges(self.id)
            for edge_x in edges:
                if edge_x.edge_type == "has_sense":
                    sense_nodes.append(CwnSense(edge_x.tgt_id, cgu))
            self._senses = sense_nodes
        return self._senses


class CwnSense(CwnAnnotationInfo):
    """Class representing a sense.
    
    Attributes
    ----------
    lemmas: list
        a list of :class:`CwnLemma <.CwnLemma>` belonging to this sense.
    relations: list
        a list of tuples ``(edge_type, end_node, edge_direction)`` 
        giving the information of the edges linked to this sense
    semantic_relations: list
        a list of senses or facets (:class:`CwnSense <.CwnSense>` or 
        :class:`CwnSynset <.CwnSynset>`) related to this sense
    hypernym: list
        a list of hypernyms (:class:`CwnSense <.CwnSense>`) of this
        sense.
    hyponym: list
        a list of hyponym (:class:`CwnSense <.CwnSense>`) of this
        sense.
    synset: list
        a synset (:class:`CwnSynset <.CwnSynset>`) containing this 
        sense.
    synonym: list
        a list of synonyms (:class:`CwnSense <.CwnSense>`) of this
        sense.
    facets: list
        a list of sense facets (:class:`CwnFacet <.CwnFacet>`) 
        belonging to this sense.
    """

    def __init__(self, nid, cgu):
        ndata = cgu.get_node_data(nid)
        self.cgu = cgu
        self.id = nid
        self.pos = ndata.get("pos", "")
        self.node_type = "sense"
        self.definition = ndata.get("def", "")
        self.src = ndata.get("src", None)
        self.examples = ndata.get("examples", [])
        self.domain = ndata.get("domain", "")
        self.annot = ndata.get("annot", {})
        self._relations = None
        self._lemmas = None

    def __repr__(self):
        try:
            head_word = self.lemmas[0].lemma
        except (IndexError, AttributeError):
            head_word = "----"
        return "<CwnSense[{id}]({head}): {definition}>".format(
            head=head_word, **self.__dict__
        )

    def __eq__(self, other):
        if isinstance(other, CwnSense):
            return self.definition == other.definition and \
                self.pos == other.pos and \
                (self.src and other.src and self.src == other.src)
        else:
            return False

    def __hash__(self):
        return hash((self.definition, self.pos, self.src))

    def data(self):
        """Retrieve all data of this sense.
        
        Returns
        -------
        dict
            data stored in a dictionary, including the following
            keys: ``node_type``, ``pos``, ``examples``, ``domain``,
            ``annot``, ``def``
        """
        data_fields = ["node_type", "pos", "examples", "domain", "annot"]
        data_dict= {
            k: self.__dict__[k] for k in data_fields
        }
        data_dict["def"] = self.definition
        return data_dict

    def all_examples(self):
        """Retrieve all example sentences of this sense, 
        including examples inside each sense facet.
        
        Returns
        -------
        list
            a list of example sentences (``str``)
        """
        examples = self.examples
        if not examples:
            examples = []
        if isinstance(examples,str) and\
            examples != '':
            examples = [examples]
        
        
        for facet_x in self.facets:
            examples.extend(facet_x.examples)
        return examples

    @property
    def lemmas(self):
        if self._lemmas is None:
            cgu = self.cgu
            lemma_nodes = []
            edges = cgu.find_edges(self.id, is_directed=False)
            for edge_x in edges:
                if edge_x.edge_type == "has_sense":
                    lemma_nodes.append(CwnLemma(edge_x.src_id, cgu))
            self._lemmas = lemma_nodes
        return self._lemmas

    @property
    def relations(self):
        if self._relations is None:
            cgu = self.cgu
            relation_infos = []
            edges = cgu.find_edges(self.id, is_directed=False)
            for edge_x in edges:
                if edge_x.edge_type.startswith("has_sense"):
                    continue                
                
                if not edge_x.reversed:
                    edge_type = edge_x.edge_type
                    end_node_id = edge_x.tgt_id
                    edge_direction = "forward"
                else:
                    edge_type = edge_x.edge_type
                    end_node_id = edge_x.src_id
                    edge_direction = "reversed"
                
                node_data = cgu.get_node_data(end_node_id) 
                if node_data.get("node_type") == "facet":
                    end_node = CwnFacet(end_node_id, cgu) 
                elif node_data.get("node_type") == "synset":
                    end_node = CwnSynset(end_node_id, cgu)
                else:
                    end_node = CwnSense(end_node_id, cgu)

                relation_infos.append((edge_type, end_node, edge_direction))

            self._relations = relation_infos
        return self._relations

    @property
    def semantic_relations(self):
        relation_infos = self.relations
        sem_relations = []
        for rel_x in relation_infos:
            rel_type = CwnRelationType[rel_x[0]]
            if rel_type.is_semantic_relation():
                sem_relations.append(rel_x)
        return sem_relations

    @property
    def hypernym(self):
        relation_infos = self.relations
        hypernym = [x[1] for x in relation_infos if x[0] == "hypernym"]
        return hypernym
    
    @property
    def hyponym(self):
        relation_infos = self.relations
        hypernym = [x[1] for x in relation_infos if x[0] == "hyponym"]
        return hypernym

    @property
    def synset(self):
        relation_infos = self.relations
        synsets = [x[1] for x in relation_infos if x[0] == "is_synset"]
        if not synsets:
            synset = None
        elif len(synsets) == 1:
            synset = synsets[0]
        elif len(synsets) > 1:
            print("WARNING: more than one synset, returning the first")
            synset = synsets[0]
        return synset

    @property
    def synonym(self):
        relation_infos = self.relations
        synonyms = [x[1] for x in relation_infos if x[0] == "synonym"]
        return synonyms

    @property
    def facets(self):
        relation_infos = self.relations
        facets = [x[1] for x in relation_infos if x[0] == "has_facet"]
        return facets

class CwnFacet(CwnSense):
    """Class representing a sense facet.

    Attributes
    ----------
    sense: CwnSense
        the sense (:class:`CwnSense <.CwnSense>`) of this 
        sense facet
    """
    def __init__(self, nid, cgu):
        super(CwnFacet, self).__init__(nid, cgu)
        self.node_type = "facet"
        self._sense = None

    def __repr__(self):
        try:
            head_word = self.sense.lemmas[0].lemma
        except (IndexError, AttributeError):
            head_word = "----"
        return "<CwnFacet[{id}]({head}): {definition}>".format(
            head=head_word, **self.__dict__
        )

    @property
    def sense(self):
        if self._sense is None:
            cgu = self.cgu            
            edges = cgu.find_edges(self.id, is_directed=False)
            for edge_x in edges:
                if edge_x.edge_type == "has_facet":
                    self._sense = CwnSense(edge_x.src_id, cgu)
                    break
        return self._sense
        

class CwnSynset(CwnAnnotationInfo):
    def __init__(self, nid, cgu):
        ndata = cgu.get_node_data(nid)
        self.cgu = cgu
        self.id = nid
        self.node_type = "synset"
        self.gloss = ndata.get("gloss", "")
        self.pwn_word = ndata.get("pwn_word", "")
        self.pwn_id = ndata.get("pwn_id", "")
        self._relations = None

    def __repr__(self):
        return "<CwnSynset[{id}]: {gloss}>".format(
            **self.__dict__
        )

    def data(self):
        data_fields = ["node_type", "gloss", "pwn_word", "pwn_id"]
        data_dict= {
            k: self.__dict__[k] for k in data_fields
        }
        return data_dict

    def __eq__(self, other):
        if isinstance(other, CwnSynset):
            return self.gloss == other.gloss
        else:
            return False

    def __hash__(self):
        return hash(self.gloss)
    
    @property
    def definition(self):
        return self.gloss

    @property
    def relations(self):
        if self._relations is None:
            cgu = self.cgu
            relation_infos = []
            edges = cgu.find_edges(self.id, is_directed=False)
            for edge_x in edges:
                if edge_x.edge_type.startswith("has_sense"):
                    continue                
                
                if not edge_x.reversed:
                    edge_type = edge_x.edge_type
                    end_node_id = edge_x.tgt_id  
                    edge_direction = "forward"                  
                else:
                    edge_type = edge_x.edge_type
                    end_node_id = edge_x.src_id
                    edge_direction = "reversed"
                
                node_data = cgu.get_node_data(end_node_id) 
                ntype = node_data.get("node_type")
                if ntype == "facet":
                    end_node = CwnFacet(end_node_id, cgu) 
                elif ntype == "sense":
                    end_node = CwnSense(end_node_id, cgu)
                elif ntype == "synset":
                    end_node = CwnSynset(end_node_id, cgu)
                else:
                    end_node = None

                relation_infos.append((edge_type, end_node, edge_direction))

            self._relations = relation_infos
        return self._relations

    @property
    def semantic_relations(self):
        relation_infos = self.relations
        sem_relations = []
        for rel_x in relation_infos:
            rel_type = CwnRelationType[rel_x[0]]
            if rel_type.is_semantic_relation():
                sem_relations.append(rel_x)
        return sem_relations

    @property
    def senses(self):
        relation_infos = self.relations
        senses = [x[1] for x in relation_infos if x[0].startswith("is_synset")]
        return senses


class CwnAnnotationInfo:
    def __init__(self):
        self.annot = {}

    @property
    def author(self):
        return self.annot.get("author", "")
    
    @author.setter
    def author(self, x):
        self.annot["author"] = x

    @property
    def confidence(self):
        return self.annot.get("confidence", 1)
    
    @confidence.setter
    def confidence(self, x):
        if not (0 < x < 1):
            raise ValueError("confidence must be 0 < x < 1")
        else:
            self.annot["confidence"] = x

    @property
    def action(self):
        return self.annot.get("action", "create")

    @action.setter
    def action(self, x):
        if x not in ("create", "update", "delete"):
            raise ValueError("action must be one of create, update, delete")
        else:
            self.annot["action"] = x


import json
import logging
import numpy as np

from transformers import BertTokenizer
from .config import BERT_MODEL
from .download import get_model_path

tokenizer = None
def warmup():
    global tokenizer
    if not tokenizer:
        tokenizer = BertTokenizer.from_pretrained(get_model_path())

def tokenize(test_data):
    all_instance = []
    logging.info("Tokenize data by %s"%(BERT_MODEL))
    
    global tokenizer
    warmup()

    for sentence in test_data:
        one_sentence = []
        for word in sentence:
            one_instance = []
            for i in word:
                if "cwn_definition" not in i:
                    one_instance.append([])
                    continue
                instance = {
                'test_word': i['test_word'],
                'test_word_id': tokenizer.convert_tokens_to_ids(tokenizer.tokenize(i['test_word'])),
                'test_sentence': i['test_sentence'],
                'test_sentence_id': tokenizer.convert_tokens_to_ids(tokenizer.tokenize(i['test_sentence'])),
                'cwn_definition': i['cwn_definition'],
                'cwn_definition_id': tokenizer.convert_tokens_to_ids(tokenizer.tokenize(i['cwn_definition'])),
                'cwn_sentence': i['cwn_sentence'],
                'cwn_sentence_id': tokenizer.convert_tokens_to_ids(tokenizer.tokenize(i['cwn_sentence'])),
                'label': bool(i['label']),
                }
                one_instance.append(instance)
            one_sentence.append(one_instance)
        all_instance.append(one_sentence) 
    logging.info("Done")
    return all_instance

if __name__ == "__main__":
    tokenize()



from .wsd import senseTag
from .preprocess import preprocess
from .tokenize_test import tokenize
from .test import test
from .cwn_graph_utils import CwnGraphUtils
from .cwn_types import GraphStructure, AnnotAction, AnnotRecord, CwnCheckerSuggestion, CwnIdNotFoundError  

from .cwn_annot_types import CwnAnnotationInfo
from .cwn_node_types import CwnNode, CwnGlyph, CwnLemma, CwnSense, CwnFacet, CwnSynset
from .cwn_relation_types import CwnRelationType, CwnRelation


from .cwn_base import CwnBase
from .config import BERT_MODEL, CLS, SEP, COMMA, PAD, USE_CUDA
from .model import WSDBertClassifer
from .util import positive_weight, accuracy
from .download import download, setup_model



 



import pickle
import hashlib
from enum import Enum, auto
from typing import Tuple
from collections import namedtuple

class AnnotAction(Enum):
    Edit = 1
    Delete = 2

class AnnotRecord:
    def __init__(self, annot_id:str, annot_action:AnnotAction, raw_id:str=None):
        self.action = annot_action
        self.annot_type = "generic"  
        self.annot_id = annot_id
        self.raw_id = raw_id

class GraphStructure:
    def __init__(self):
        self.V = {}
        self.E = {}
        self.meta = {}
        self._hash = None

    def compute_dict_hash(self, dict_obj):        
        m = hashlib.sha1()
        for k, value in sorted(dict_obj.items()):
            if isinstance(value, dict):
                m.update(pickle.dumps(k))                
                value_hash = self.compute_dict_hash(value)
                m.update(value_hash.encode())                 
            else:
                m.update(pickle.dumps((k, value)))                                       
        hash_value = m.hexdigest()
        return hash_value

    def get_hash(self):
        if not self._hash:
            Vhash = self.compute_dict_hash(self.V)
            Ehash = self.compute_dict_hash(self.E)
            m = hashlib.sha1()
            m.update(Vhash.encode())
            m.update(Ehash.encode())
            self._hash = m.hexdigest()        
        hashStr = self._hash[:6]
        return hashStr

    def export(self):
        print("export Graph ", self.get_hash())
        print("export to cwn_graph.pyobj, "
              "you may need to install it with CwnBase.install_cwn()")
        with open("data/cwn_graph.pyobj", "wb") as fout:
            pickle.dump((self.V, self.E), fout)

class CwnCheckerSuggestion(Enum):
    MISSING_SYNSET = auto()
    NO_SYNSET = auto()
    SYN_NO_SENSE = auto()
    SYN_WRONG_DEF = auto()
    SYN_MISSING_REL = auto()
    SYN_REL_DIFF = auto()
    INVERSE_ERROR = auto()
    INVERSE_NOT_EXISTS = auto()
csg = CwnCheckerSuggestion

SuggestionData = Tuple[CwnCheckerSuggestion, any]
class CwnIdNotFoundError(Exception):
    pass



import logging
import numpy as np
import os
import torch
import warnings

from argparse import ArgumentParser
from transformers import AdamW, get_linear_schedule_with_warmup
from torch.utils.data import DataLoader, random_split

from .config import CLS, SEP, COMMA, PAD
from . import config
from .model import WSDBertClassifer
from .util import positive_weight, accuracy
from .download import get_model_path

wsd_model = None

def batch_generation(batch_size, data):
    idx = 0
    target_idx = min(len(data), idx+batch_size)
    all_batch = []
    while idx < len(data):
        target_idx = min(len(data), idx+batch_size)
        max_length = min(max([len(data[i]['test_sentence_id'])+len(data[i]['test_word_id']) + len(data[i]['cwn_definition_id']) + len(data[i]['cwn_sentence_id']) + 5 for i in range(idx, target_idx) ]), 512)
        context = []
        attention_mask = []
        token_type_id = []
        label = []
        for i in range(idx, target_idx):
            attention = np.ones(max_length)
            token_type = np.zeros(max_length)
            text = [CLS] + data[i]['test_sentence_id'] + [SEP] + data[i]['test_word_id'] + [COMMA] + data[i]['cwn_definition_id'] + [COMMA] + data[i]['cwn_sentence_id'] + [SEP]

            pad_length = max(max_length-len(text),0)
            token_type[len(data[i]['test_sentence_id'])+2:len(text)] = 1
            if pad_length > 0:
                attention[-pad_length:] = 0
                text += [PAD for i in range(pad_length)]
            context.append(text)
            attention_mask.append(attention)
            token_type_id.append(token_type)
            label.append(data[i]['label'])
        batch = {
            "context": torch.tensor(context),
            "attention_mask": torch.tensor(attention_mask),
            "token_type_ids": torch.tensor(token_type_id),
            "label": torch.tensor(label)
            }
        all_batch.append(batch)
        idx += batch_size

    return all_batch

def warmup():
    global wsd_model
    if not wsd_model:
        device = torch.device('cuda' if torch.cuda.is_available() and config.USE_CUDA else 'cpu')
        logging.info('Device type is %s'%(device))
        logging.info("Prepare Dataset")
        wsd_model = WSDBertClassifer.from_pretrained(get_model_path())
        wsd_model.to(device)
        wsd_model.eval()

@torch.no_grad()
def test(all_json, batch_size=8):
    global wsd_model

    device = torch.device('cuda' if torch.cuda.is_available() and config.USE_CUDA else 'cpu')
    warnings.filterwarnings("ignore")
    warmup()

    # model.load_state_dict(checkpoint['state_dict'])
    all_ans = []
    logging.info("Start Inference")
    for sentence in all_json:
        sentence_ans =[]
        for word in sentence:
            if word[0] == []:
                sentence_ans.append(-1)
                continue

            batches = batch_generation(batch_size, word)
            one_label = []
            one_predict = []
            for b in batches:
                context = b['context'].to(device)
                attention_mask = b['attention_mask'].to(device)
                token_type_ids = b['token_type_ids'].to(torch.long).to(device)

                logits, _ = wsd_model(context, attention_mask=attention_mask, token_type_ids=token_type_ids)
                one_predict += logits.squeeze(1).tolist()

            sentence_ans.append(np.argmax(one_predict))
        all_ans.append(sentence_ans)
    logging.info("Done")
    return all_ans




import argparse
import csv
import json
import logging
import os

from .cwn_base import CwnBase
from .download import get_model_path

__cwn_inst = None
__cwn_cache = {}

def simplify_pos(pos):
    relation = {
        "A": "OTHER",
        "Caa": "OTHER",
        "Cab": "OTHER",
        "Cba": "OTHER",
        "Cbb": "OTHER",
        "Da": "OTHER",
        "Dfa": "OTHER",
        "Dfb": "OTHER",
        "Di": "OTHER",
        "Dk": "OTHER",
        "D": "OTHER",
        "Na": "N",
        "Nb": "Nb",
        "Nc": "N",
        "Ncd": "N",
        "Nd": "N",
        "Neu": "OTHER",
        "Nes": "OTHER",
        "Nep": "OTHER",
        "Neqa": "OTHER",
        "Neqb": "OTHER",
        "Nf": "OTHER",
        "Ng": "OTHER",
        "Nh": "N",
        "Nv": "N",
        "I": "OTHER",
        "P": "OTHER",
        "T": "OTHER",
        "nom,VA": "V",
        "VA": "V",
        "VAC": "V",
        "VB": "V",
        "VC": "V",
        "VCL": "V",
        "VD": "V",
        "VE": "V",
        "VF": "V",
        "VG": "V",
        "VH": "V",
        "VHC": "V",
        "VI": "V",
        "VJ": "V",
        "VK": "V",
        "VL": "V",
        "COLONCATEGORY": "OTHER",
        "COMMACATEGORY": "OTHER",
        "DASHCATEGORY": "OTHER",
        "ETCCATEGORY": "OTHER",
        "EXCLAMATIONCATEGORY": "OTHER",
        "PARENTHESISCATEGORY": "OTHER",
        "PAUSECATEGORY": "OTHER",
        "PERIODCATEGORY": "OTHER",
        "QUESTIONCATEGORY": "OTHER",
        "SEMICOLONCATEGORY": "OTHER",
        "SPCHANGECATEGORY": "OTHER",
    }
    try:
        return relation[pos]
    except KeyError as e:
        return "OTHER"


#Search and filter senses from CWN
def get_cwn_senses(cwn, word_info):
    word = word_info["word"]
    if word in __cwn_cache:
        senses = __cwn_cache[word]
    else:
        senses = cwn.find_all_senses(word)
        __cwn_cache[word] = senses

    if len(senses) == 0:
        return []
    elif word_info['pos'] == "":
        return senses
    else:
        simple_pos = simplify_pos(word_info["pos"])
        same_pos_senses = list(filter(lambda x:simplify_pos(x.pos)==simple_pos, senses))
        return same_pos_senses


#Generate original sentence with tag on target word
def generate_sentence(target, sentence_info):
    sentence = ""
    for i in range(len(sentence_info)):
        if i == target:
            sentence += "<"+sentence_info[i][0]+">"
        else:
            sentence += sentence_info[i][0]
    return sentence


#Generate a row of training data
def generate_row(word_info, cwn_sense):
    row = {}
    row['test_word'] = word_info["word"]
    row['test_pos'] = word_info["pos"]
    row['test_sense_id'] = word_info["sense_id"]
    row['test_definition'] = word_info["definition"]
    row['test_sentence'] = word_info["sentence"]
    row['cwn_sense_id'] = cwn_sense.id
    row['cwn_definition'] = cwn_sense.definition
    tmp = [x for x in cwn_sense.all_examples() if x.strip()]
    if len(tmp) > 0:
        row['cwn_sentence'] = tmp[0]
    else:
        return None
    row['label'] = True if word_info["sense_id"] == cwn_sense.id else False
    return row

def get_cwn_inst():
    global __cwn_inst
    if not __cwn_inst:
        logging.info("Preprocessing data")        
        logging.info("loading CWN")        
        __cwn_inst = CwnBase()
        logging.info("Done")
    return __cwn_inst

# data: a list of sentences
def preprocess(data):
    
    all_data = []
    cwn = get_cwn_inst()
    logging.info("Query CWN graph to build test data")

    for sentence_info in data:
        one_sentence = []
        for idx, i in enumerate(sentence_info):
            one_instance = []
            word_info = {}
            word_info["word"] = i[0]
            word_info["pos"] = i[1]
            word_info["sense_id"] = ""
            word_info["definition"] = ""
            
            word_info["sentence"] = generate_sentence(idx, sentence_info)                          
            cwn_senses = get_cwn_senses(cwn, word_info)
            
            if len(cwn_senses) == 0:
                one_instance.append({'test_word': word_info["word"], 'test_pos' :word_info["pos"]} )
            for sense in cwn_senses:
                row = generate_row(word_info, sense)
                if row is not None:
                    one_instance.append(row)
            one_sentence.append(one_instance)
        all_data.append(one_sentence)
    logging.info("Done")
    return all_data




from .preprocess import preprocess
from .tokenize_test import tokenize
from .test import test

def senseTag(data):
    cwn_batch = preprocess(data)
    tokenize_data = tokenize(cwn_batch)
    ans_idx = test(tokenize_data)

    all_ans = []
    for batch, sentence_idx  in zip(cwn_batch, ans_idx):
        sentence_ans = []
        for word, idx in zip(batch, sentence_idx):
            if idx == -1:
                word_ans = [word[idx]['test_word'], word[idx]['test_pos'], "", ""]
            else:
                word_ans = [word[idx]['test_word'], word[idx]['test_pos'], word[idx]['cwn_sense_id'], word[idx]['cwn_definition']]
            sentence_ans.append(word_ans)
        all_ans.append(sentence_ans)
    return all_ans



import os
import zipfile
import tempfile

def get_model_path():
    home_dir = os.path.expanduser("~")
    cache_dir = os.path.join(home_dir, ".cwn_graph")
    model_path = os.path.join(cache_dir, "cwn-wsd-model")
    return model_path

def download(upgrade=False):
    if os.path.exists(get_model_path()) and not upgrade:
        print("A copy of DistilTag model already exists. Use upgrade=True to overwrite")
        return 

    import gdown    
    url = "https://drive.google.com/uc?id=1xTdrqOuvLFs2ElHUmCJIcrnO3fKrbH5K"
    with tempfile.TemporaryDirectory("distiltag") as tmpdir:
        outpath = os.path.join(tmpdir, "cwn-wsd-model.zip")
        gdown.download(url, outpath, quiet=False)        
        setup_model(outpath)

def setup_model(zip_path):
    print("setting up model...")
    home_dir = os.path.expanduser("~")
    cache_dir = os.path.join(home_dir, ".cwn_graph")
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir, exist_ok=True)

    with zipfile.ZipFile(zip_path, "r") as ref:
        ref.extractall(cache_dir)

    print("CwnSenseTagger model installed.")


import pickle
import os
from .download import get_model_path
from .cwn_graph_utils import CwnGraphUtils

class CwnBase(CwnGraphUtils):
    """The base cwn reference data.
    """

    def __init__(self):        
        model_path = get_model_path()
        cwn_path = os.path.join(model_path, "cwn_graph.pyobj")                   
        if not os.path.exists(cwn_path):
            raise FileNotFoundError("Cannot find model data, have you tried CwnSenseTagger.download() ?")

        with open(cwn_path, "rb") as fin:
            data = pickle.load(fin)
            if len(data) == 2:
                V, E = data
                meta = {}
            else:
                V, E, meta = data
        super(CwnBase, self).__init__(V, E, meta)        
    


import os
import logging
import numpy as np


def positive_weight(train_generator):
    all_instance = []
    for i in train_generator:
        all_instance += i['label'].tolist()
    pos = sum(all_instance)
    neg = len(all_instance)-pos
    return neg / pos

def accuracy(label, predict):
    correct = ~(np.array(label)^np.array(predict))+2
    return correct.sum()/len(correct)



import re
from itertools import chain
from .cwn_types import GraphStructure, AnnotAction, AnnotRecord, CwnCheckerSuggestion, CwnIdNotFoundError  
from .cwn_node_types import CwnNode, CwnGlyph, CwnLemma, CwnSense, CwnFacet, CwnSynset
from .cwn_relation_types import CwnRelationType, CwnRelation


class CwnGraphUtils(GraphStructure):
    """cwn data as graph (vertices and edges)
    """

    def __init__(self, V, E, meta={}):
        super(CwnGraphUtils, self).__init__()
        self.V = V
        self.E = E
        self.meta = meta
        self.edge_src_index = self.build_index(E.keys(), lambda x: x[0])
        self.edge_tgt_index = self.build_index(E.keys(), lambda x: x[1])

    def build_index(self, data, keyfunc):
        idx = {}        
        for k in data:                   
            idx_key = keyfunc(k)
            idx.setdefault(idx_key, []).append(k)
        return idx

    def find_glyph(self, instr):
        for v, vdata in self.V.items():
            if vdata["node_type"] == "glyph":
                if vdata["glyph"] == instr:               
                   return v
        return None
    
    def find_lemma(self, instr_regex):
        """Find lemmas matching search pattern.
        
        Parameters
        ----------
        instr_regex : str
            RegEx pattern to search for.
        
        Returns
        -------
        list
            A list of :class:`CwnLemma <CwnGraph.cwn_types.CwnLemma>`.
        """
        ret = []
        pat = re.compile(instr_regex)
        for v, vdata in self.V.items():
            if vdata["node_type"] == "lemma":
                if pat.search(vdata["lemma"]) is not None:               
                   ret.append(CwnLemma(v, self))
        return ret

    def find_senses(self, lemma="", definition="", examples=""):
        """Find senses with lemmas, definitions, or examples matching 
        search patterns.
        
        Parameters
        ----------
        lemma : str, optional
            RegEx pattern for searching the lemma of a sense, by default ""
        definition : str, optional
            RegEx pattern for searching the definition of a sense, by default ""
        examples : str, optional
            RegEx pattern for searching the examples of a sense, by default ""
        
        Returns
        -------
        list
            A list of :class:`CwnSense <CwnGraph.cwn_types.CwnSense>` matching 
        """

        lemma_re = re.compile(lemma)
        def_re = re.compile(definition)
        ex_re = re.compile(examples)

        sense_list = []
        for node_id, node_x in self.V.items():
            if not node_x["node_type"] == "sense":
                continue
            sense_x = CwnSense(node_id, self)            
            if lemma:
                lemma_matched = any(lemma_re.search(lemma_x.lemma) 
                    for lemma_x in sense_x.lemmas)
            else:
                lemma_matched = False
            
            if definition:
                def_matched = def_re.search(sense_x.definition)
            else:
                def_matched = False
            
            if examples:
                ex_san_re = re.compile(r"[\<\>]")
                example_matched = any([ex_re.search(ex_san_re.sub("", ex_x))
                    for ex_x in sense_x.examples])
            else:
                example_matched = False

            if lemma_matched or def_matched or \
                example_matched:
                sense_list.append(sense_x)                
        return sense_list            

    def find_all_senses(self, lemma):    
        sense_iter = (x.senses for x in self.find_lemma(f"^{lemma}$"))
        sense_iter = chain.from_iterable(sense_iter)
        return list(sense_iter)       

    def find_edges(self, node_id, is_directed = True):
        ret = []
        
        for e in self.edge_src_index.get(node_id, []):  
            ret.append(CwnRelation(e, self))            
        if not is_directed:
            for e in self.edge_tgt_index.get(node_id, []):
                ret.append(CwnRelation(e, self, reversed=True))   

        return ret
    
    def connected(self, node_id, is_directed = True, maxConn=100, sense_only=True):
        ret = []
        visited = set()
        buf = [node_id]
        while buf:
            node_x = buf.pop()
            visited.add(node_x)
            conn_edges = self.find_edges(node_x, is_directed)            
            for conn_edge_x in conn_edges:
                conn_node_x = conn_edge_x[0]
                conn_rel = conn_edge_x[1]
                if sense_only and "has_sense" in conn_rel:
                    continue

                if conn_node_x in visited or conn_node_x in buf:
                    continue
                else:
                    buf.append(conn_node_x)
                ret.append((node_x, conn_rel, conn_node_x))
            if maxConn and len(ret) > maxConn:
                break                        
        return ret       

    def has_id(self, node_id):
        return node_id in self.V or node_id in self.E
        
    def get_node_data(self, node_id, field_name = None):
        return self.V.get(node_id, {})

    def get_edge_data(self, edge_id, field_name = None):
        return self.E.get(edge_id, {})
        
    def from_sense_id(self, sense_id):
        return CwnSense(sense_id, self)


import setuptools

with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="CwnSenseTagger", 
    version="0.1.5",
    author="NTUGIL LOPE Lab",    
    description="A package to use chinese word net to achieve word sense disambigution task",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",        
        "Operating System :: OS Independent",
    ],
    setup_requires=["wheel"],
    install_requires=[
        "numpy",        
        "pandas",        
        "torch>=1.6",
        "tqdm",
        "transformers>=3.2",
    ],
    include_package_data=True,
    python_requires='>=3.5',
)
